library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity axi_tb is end axi_tb;

architecture testbench of axi_tb is
	signal clk, enable : std_logic;

    signal awid    : std_logic_vector(5 downto 0);
    signal awaddr  : std_logic_vector(31 downto 0);
    signal awlen   : std_logic_vector(3 downto 0);
    signal awsize  : std_logic_vector(1 downto 0);
    signal awburst : std_logic_vector(1 downto 0);
    signal awlock  : std_logic_vector(1 downto 0);
    signal awcache : std_logic_vector(3 downto 0);
    signal awprot  : std_logic_vector(2 downto 0);
    signal awvalid : std_logic;
    signal awready : std_logic;

    signal wid     : std_logic_vector(5 downto 0);
    signal wdata   : std_logic_vector(63 downto 0);
    signal wstrb   : std_logic_vector(7 downto 0);
    signal wlast   : std_logic;
    signal wvalid  : std_logic;
    signal wready  : std_logic;

    signal bid     : std_logic_vector(5 downto 0);
    signal bresp   : std_logic_vector(1 downto 0);
    signal bvalid  : std_logic;
    signal bready  : std_logic;

    signal arid    : std_logic_vector(5 downto 0);
    signal araddr  : std_logic_vector(31 downto 0);
    signal arlen   : std_logic_vector(3 downto 0);
    signal arsize  : std_logic_vector(1 downto 0);
    signal arburst : std_logic_vector(1 downto 0);
    signal arlock  : std_logic_vector(1 downto 0);
    signal arcache : std_logic_vector(3 downto 0);
    signal arprot  : std_logic_vector(2 downto 0);
    signal arvalid : std_logic;
    signal arready : std_logic;

    signal rid     : std_logic_vector(5 downto 0);
    signal rdata   : std_logic_vector(63 downto 0);
    signal rresp   : std_logic_vector(1 downto 0);
    signal rlast   : std_logic;
    signal rvalid  : std_logic;
    signal rready  : std_logic;

	constant clk_period : time := 30 ns;

    signal dwready : std_logic;
    signal dawready : std_logic;
    signal dbvalid : std_logic;

    signal drvalid : std_logic;
    signal darready : std_logic;
    signal debug : std_logic;
begin
SLAVE: entity work.axi_slave
	port map (
        aclk    => clk,
        aresetn => not enable,

        awid    => awid,
        awaddr  => awaddr,
        awlen   => awlen,
        awsize  => awsize,
        awburst => awburst,
        awlock  => awlock,
        awcache => awcache,
        awprot  => awprot,
        awvalid => awvalid,
        awready => awready,

        wid     => wid   ,
        wdata   => wdata ,
        wstrb   => wstrb ,
        wlast   => wlast ,
        wvalid  => wvalid,
        wready  => wready,

        bid     => bid   ,
        bresp   => bresp ,
        bvalid  => bvalid,
        bready  => bready,

        arid    => arid  ,
        araddr  => araddr,
        arlen   => arlen ,
        arsize  => arsize,
        arburst => arburst,
        arlock  => arlock,
        arcache => arcache,
        arprot  => arprot,
        arvalid => arvalid,
        arready => arready,

        rid     => rid   ,
        rdata   => rdata ,
        rresp   => rresp ,
        rlast   => rlast ,
        rvalid  => rvalid,
        rready  => rready,

		debug => debug
	);

    dwready <= wready after 2ns;
    dawready <= awready after 2ns;
    dbvalid <= bvalid after 2ns;

    darready <= arready after 2ns;
    drvalid <= rvalid after 2ns;

	-- Clock process
	clk_proc : process
	begin
		clk <= '1';
		wait for clk_period / 2;
		clk <= '0';
		wait for clk_period / 2;
	end process;

    stim_proc : process
    begin
        enable <= '0';

        awvalid <= '0';
        wvalid <= '0';
        bready <= '0';

        arvalid <= '0';
        rready <= '0';

        wait for clk_period * 4;
        enable <= '1';
        wait for clk_period * 4;
        wait for clk_period / 4;

        awaddr <= X"00000000";
        awvalid <= '1';
        awsize <= b"10";

        while 1 = 1 loop
            wait until clk = '1';
            exit when dawready = '1';
        end loop;
        wait for clk_period / 4;

        awvalid <= '0';

        wdata <= X"0123456789abcdef";
        wstrb <= X"0f";
        wvalid <= '1';
        wlast <= '0';

        while 1 = 1 loop
            wait until clk = '1';
            exit when dwready = '1';
        end loop;
        wait for clk_period / 4;

        wstrb <= X"f0";
        wlast <= '1';

        while 1 = 1 loop
            wait until clk = '1';
            exit when dwready = '1';
        end loop;
        wait for clk_period / 4;

        wvalid <= '0';

        while 1 = 1 loop
            wait until clk = '1';
            exit when dbvalid = '1';
        end loop;
        wait for clk_period / 4;

        bready <= '1';
        wait until clk = '1';
        bready <= '0';

        wait for clk_period / 4;
        araddr <= X"00000000";
        arvalid <= '1';
        arsize <= b"10";
        arlen  <= X"1";

        while 1 = 1 loop
            wait until clk = '1';
            exit when darready = '1';
        end loop;
        wait for clk_period / 4;
        arvalid <= '0';

        while 1 = 1 loop
            wait until clk = '1';
            exit when drvalid = '1';
        end loop;
        wait for clk_period / 4;
        rready <= '1';
        wait until clk = '1';
        wait for clk_period / 4;
        rready <= '0';

        while 1 = 1 loop
            wait until clk = '1';
            exit when drvalid = '1';
        end loop;
        wait for clk_period / 4;
        rready <= '1';
        wait until clk = '1';
        wait for clk_period / 4;
        rready <= '0';


        wait;
    end process;

end testbench;
