library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity toplevel_tb is end toplevel_tb;

architecture testbench of toplevel_tb is
	signal clk, enable : std_logic;
    signal xu              : std_logic_vector(143 downto 0);
    signal xl              : std_logic_vector(143 downto 0);

    signal yu              : std_logic_vector(143 downto 0);

    signal m_awid          : std_logic_vector(1 downto 0);
    signal m_awaddr        : std_logic_vector(15 downto 0);
    signal m_awlen         : std_logic_vector(1 downto 0);
    signal m_awsize        : std_logic;
    signal m_awburst       : std_logic;
    signal m_awlock        : std_logic;
    signal m_awcache       : std_logic_vector(1 downto 0);
    signal m_awprot        : std_logic_vector(1 downto 0);
    signal m_awvalid       : std_logic;
    signal awready         : std_logic;

    signal m_wid           : std_logic_vector(1 downto 0);
    signal m_wdata         : std_logic_vector(31 downto 0);
    signal m_wstrb         : std_logic_vector(3 downto 0);
    signal m_wlast         : std_logic;
    signal m_wvalid        : std_logic;
    signal wready          : std_logic;

    signal m_bid           : std_logic_vector(2 downto 0);
    signal m_bresp         : std_logic;
    signal m_bvalid        : std_logic;
    signal bready          : std_logic;

    signal m_arid          : std_logic_vector(1 downto 0);
    signal m_araddr        : std_logic_vector(15 downto 0);
    signal m_arlen         : std_logic_vector(1 downto 0);
    signal m_arsize        : std_logic;
    signal m_arburst       : std_logic;
    signal m_arlock        : std_logic;
    signal m_arcache       : std_logic_vector(1 downto 0);
    signal m_arprot        : std_logic_vector(1 downto 0);
    signal m_arvalid       : std_logic;
    signal arready         : std_logic;

    signal m_rid           : std_logic_vector(1 downto 0);
    signal m_rdata         : std_logic_vector(31 downto 0);
    signal m_rresp         : std_logic;
    signal m_rlast         : std_logic;
    signal m_rvalid        : std_logic;
    signal rready          : std_logic;

    signal z               : std_logic_vector(127 downto 0);
    signal zu              : std_logic_vector(234 downto 128);
    signal zl              : std_logic_vector(233 downto 200);

    signal fnce            : std_logic;

    signal nxl_fold        : std_logic;
    signal nxu_fold        : std_logic;
    signal nx_thru         : std_logic;

    signal nyl_fold        : std_logic;
    signal nyu_fold        : std_logic;
    signal ny_thru         : std_logic;

    signal nzl_fold        : std_logic;
    signal nzu_fold        : std_logic;
    signal nz_thru         : std_logic;

    signal user_sw         : std_logic_vector(7 downto 0);
    signal user_led        : std_logic_vector(7 downto 0);
    signal npb             : std_logic;

    signal nrtcken         : std_logic;

    signal clk_glbl_nen    : std_logic;

    signal d_tck, d_tdi    : std_logic;
    signal d_tdo, d_rtck   : std_logic;

    signal nsyspor         : std_logic;
    signal nsysrst         : std_logic;

    signal clk_neg_dn_out  : std_logic;
    signal clk_neg_up_out  : std_logic;
    signal clk_pos_dn_out  : std_logic;
    signal clk_pos_up_out  : std_logic;

    signal clk_neg_dn_in   : std_logic;
    signal clk_neg_up_in   : std_logic;
    signal clk_pos_dn_in   : std_logic;
    signal clk_pos_up_in   : std_logic;

    signal clk0            : std_logic;
    signal clk1            : std_logic;
    signal clk2            : std_logic;
    signal clk_extern      : std_logic;
    signal clk_buf_loop    : std_logic;
    signal clk_24mhz_fpga  : std_logic;
    signal clk_global_out  : std_logic;
    signal clk_in_minus1   : std_logic;
    signal clk_loop0_in    : std_logic;
    signal clk_loop1_out   : std_logic;
    signal clk_loop0_out   : std_logic;
    signal clk_loop1_in    : std_logic;
    signal clk_out_to_buf  : std_logic;


	constant clk_period : time := 40 ns;

begin
    TOP : entity work.toplevel
    port map (
        xu              => xu            ,
        xl              => xl            ,
        yu              => yu            ,

        m_awid          => m_awid    ,
        m_awaddr        => m_awaddr  ,
        m_awlen         => m_awlen   ,
        m_awsize        => m_awsize  ,
        m_awburst       => m_awburst ,
        m_awlock        => m_awlock  ,
        m_awcache       => m_awcache ,
        m_awprot        => m_awprot  ,
        m_awvalid       => m_awvalid ,
        awready         => awready   ,

        m_wid           => m_wid     ,
        m_wdata         => m_wdata   ,
        m_wstrb         => m_wstrb   ,
        m_wlast         => m_wlast   ,
        m_wvalid        => m_wvalid  ,
        wready          => wready    ,

        m_bid           => m_bid     ,
        m_bresp         => m_bresp   ,
        m_bvalid        => m_bvalid  ,
        bready          => bready    ,

        m_arid          => m_arid    ,
        m_araddr        => m_araddr  ,
        m_arlen         => m_arlen   ,
        m_arsize        => m_arsize  ,
        m_arburst       => m_arburst ,
        m_arlock        => m_arlock  ,
        m_arcache       => m_arcache ,
        m_arprot        => m_arprot  ,
        m_arvalid       => m_arvalid ,
        arready         => arready   ,

        m_rid           => m_rid     ,
        m_rdata         => m_rdata   ,
        m_rresp         => m_rresp   ,
        m_rlast         => m_rlast   ,
        m_rvalid        => m_rvalid  ,
        rready          => rready    ,

        z               => z             ,
        zu              => zu            ,
        zl              => zl            ,
        fnce            => fnce          ,
        nxl_fold        => nxl_fold      ,
        nxu_fold        => nxu_fold      ,
        nx_thru         => nx_thru       ,
        nyl_fold        => nyl_fold      ,
        nyu_fold        => nyu_fold      ,
        ny_thru         => ny_thru       ,
        nzl_fold        => nzl_fold      ,
        nzu_fold        => nzu_fold      ,
        nz_thru         => nz_thru       ,
        user_sw         => user_sw       ,
        user_led        => user_led      ,
        npb             => npb           ,
        nrtcken         => nrtcken       ,
        clk_glbl_nen    => clk_glbl_nen  ,
        d_tck => d_tck                   ,
        d_tdi => d_tdi                   ,
        d_tdo => d_tdo                   ,
        d_rtck => d_rtck                 ,
        nsyspor         => nsyspor       ,
        nsysrst         => nsysrst       ,
        clk_neg_dn_out  => clk_neg_dn_out,
        clk_neg_up_out  => clk_neg_up_out,
        clk_pos_dn_out  => clk_pos_dn_out,
        clk_pos_up_out  => clk_pos_up_out,
        clk_neg_dn_in   => clk_neg_dn_in ,
        clk_neg_up_in   => clk_neg_up_in ,
        clk_pos_dn_in   => clk_pos_dn_in ,
        clk_pos_up_in   => clk_pos_up_in ,
        clk0            => clk0          ,
        clk1            => clk1          ,
        clk2            => clk2          ,
        clk_extern      => clk_extern    ,
        clk_buf_loop    => clk_buf_loop  ,
        clk_24mhz_fpga  => clk_24mhz_fpga,
        clk_global_out  => clk_global_out,
        clk_in_minus1   => clk_in_minus1 ,
        clk_loop0_in    => clk_loop0_in  ,
        clk_loop1_out   => clk_loop1_out ,
        clk_loop0_out   => clk_loop0_out ,
        clk_loop1_in    => clk_loop1_in  ,
        clk_out_to_buf  => clk_out_to_buf
    );

    clk_in_minus1 <= clk;

	-- Clock process
	clk_proc : process
	begin
		clk <= '1';
		wait for clk_period / 2;
		clk <= '0';
		wait for clk_period / 2;
	end process;

    stim_proc : process
	    variable done : boolean := false;
    begin
	    m_awvalid <= '0';
	    m_arvalid <= '0';
        nsysrst <= '0';

        wait for clk_period * 4;
        nsysrst <= '1';
        wait for clk_period * 20;

		-------- writing ----------
        -- Write address

--		wait until clk = '1';
--		while not done loop
			wait for clk_period / 8;
			m_awaddr <= x"1234";
			m_awid   <= b"10";
			m_awlen  <= b"00";
			m_awvalid <= '1';
			m_awprot(1) <= '0';


			wait until clk = '0';
			wait for clk_period / 8;
			m_awaddr <= x"5678";
			m_awid   <= b"01";
			m_awlen  <= b"00";
			m_awvalid <= '1';
			m_awprot(1) <= '0';

			wait until clk = '1';
--			done := awready = '1';
--		end loop;

	    m_awvalid <= '0';

		while not done loop
			wait for clk_period / 8;
			m_wdata <= x"12345678";
			m_wvalid <= '1';
			m_wlast <= '1';
			m_wstrb <= x"f";


			wait until clk = '0';
			wait for clk_period / 8;
			m_wdata <= x"87654321";
			m_wvalid <= '0';
			m_wlast <= '0';
			m_wstrb <= x"0";

			wait until clk = '1';
			done := wready = '1';
		end loop;


        wait for clk_period / 8;
	    m_wvalid  <= '0';
        bready <= '1';
        wait until m_bvalid = '1';
        wait until clk = '1';
        wait for clk_period / 8;

        -------- reading ----------

        -- Read address data
        wait until clk = '1';
        wait for clk_period / 8;
        m_araddr <= x"1234";
        m_arid   <= b"10";
        m_arlen  <= b"00";
        m_arvalid <= '1';
        m_arprot(1) <= '0';

        wait until clk = '0';
        wait for clk_period / 8;
        m_araddr <= x"5678";
        m_arid   <= b"01";
        m_arlen  <= b"00";
        m_arvalid <= '1';
        m_arprot(1) <= '0';

        wait until clk = '1';
        wait for clk_period / 8;
	    m_arvalid <= '0';
        rready <= '1';
        wait until m_rvalid = '1';
        wait until clk = '1';
        wait for clk_period / 8;
--        rready <= '0';

        wait;
    end process;

end testbench;
