library ieee;
use ieee.std_logic_1164.all;
use work.axi.all;

entity axi_regslice is
    port (
        clk    : in std_logic;
        nreset : in std_logic;

        -- Write response channel
        s_b       : in  axi_write_resp_channel;
        s_bready  : out std_logic;
        m_b       : out axi_write_resp_channel;
        m_bready  : in std_logic;

        -- Read data channel
        s_r       : in  axi_read_data_channel;
        s_rready  : out std_logic;
        m_r       : out  axi_read_data_channel;
        m_rready  : in  std_logic
    );
end axi_regslice;

architecture axi_regslice_arch of axi_regslice is
    -- Registers
    signal b : axi_write_resp_channel;
    signal r : axi_read_data_channel;

    signal bfull : std_logic;
    signal rfull : std_logic;

begin
    process (clk, nreset) is
    begin
        if nreset = '0' then
            rfull <= '0';
            bfull <= '0';
        elsif rising_edge(clk) then
			bfull <= ((not bfull) and s_b.valid) or (bfull and (not m_bready));

            if ((not bfull) and s_b.valid) = '1' then
                b.id   <= s_b.id;
                b.resp <= s_b.resp;
            end if;

			rfull <= ((not rfull) and s_r.valid) or (rfull and (not m_rready));

			if ((not rfull) and s_r.valid) = '1' then
                r.id      <= s_r.id;
                r.data    <= s_r.data;
                r.resp    <= s_r.resp;
                r.last    <= s_r.last;
            end if;
        end if;
    end process;

    m_b.id    <= b.id;
    m_b.resp  <= b.resp;
    m_b.valid <= bfull;

    m_r.id    <= r.id;
    m_r.data  <= r.data;
    m_r.resp  <= r.resp;
    m_r.last  <= r.last;
    m_r.valid <= rfull;

    s_rready <= not rfull;
    s_bready <= not bfull;
end axi_regslice_arch;
