library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.ca_pkg.all;
use work.fitness_pkg.all;

entity ca is
    generic (
        width  : integer := 8;
        height : integer := 8
    );
    port (
        clk    : in  std_logic;
        reset  : in  std_logic;

        i      : in  std_logic_vector(height - 1 downto 0);
        o      : out std_logic_vector(height - 1 downto 0);
        evolve : in  std_logic;
        best   : out natural;
        func   : in  fitness_function
    );
end ca;

architecture ca_arch of ca is
    type wire_array is array (width + 1 downto 0, height - 1 downto 0) of ca_interface;

    signal output    : std_logic_vector(height - 1 downto 0);
    signal input     : std_logic_vector(height - 1 downto 0);
    signal ea_input  : std_logic_vector(height - 1 downto 0);
    signal wires     : wire_array;
    signal good, bad : std_logic;
begin
    o <= output;
    input <= i when evolve = '0' else ea_input;


    GEN_Y:
    for y in 0 to height - 1 generate
        output(y) <= wires(width, y).o;

        wires(width + 1, y).o <= '0';
        wires(width + 1, y).l <= x"00000000";

        wires(0,         y).o <= input(y);
        wires(0,         y).l <= x"00000000";

        GEN_X :
        for x in 1 to width generate
            CELL : entity work.ca_cell
            generic map (
                lfsr_seed => x + y * height
            )
            port map (
                clk    => clk,
                reset  => reset,
                evolve => evolve,
                good   => good,
                bad    => bad,
                n      => wires(x, (y - 1 + height) mod height),
                s      => wires(x, (y + 1)          mod height),
                e      => wires((x + 1), y),
                w      => wires((x - 1), y),
                o      => wires(x, y)
            );
        end generate;
    end generate;

    EA : entity work.ea
    generic map (
        width => height
    )
    port map (
        clk    => clk,
        reset  => reset,
        ca_out => output,
        ca_in  => ea_input,
        best   => best,
        func   => func,
        good   => good,
        bad    => bad
    );
end ca_arch;
