library ieee;
use ieee.std_logic_1164.all;
use work.fitness_pkg.all;

entity ea is
    generic (
        width : integer := 8
    );
    port (
        clk    : in  std_logic;
        reset  : in  std_logic;

        ca_out : in  std_logic_vector(width - 1 downto 0);
        ca_in  : out std_logic_vector(width - 1 downto 0);

        best   : out natural;
        func   : in  fitness_function;

        good   : out std_logic;
        bad    : out std_logic
    );
end ea;

architecture ea_arch of ea is
    function count_ones(v : std_logic_vector) return natural is
        variable cnt : natural := 0;
    begin
        for i in v'low to v'high loop
            if v(i) = '1' then
                cnt := cnt + 1;
            end if;
        end loop;
        return cnt;
    end function;

    signal lfsr_o : std_logic_vector(31 downto 0);

    signal i,j : natural;
    signal bst : natural;
    signal cur : natural;
    signal okay: std_logic_vector(width - 1 downto 0);
    signal input  : std_logic_vector(width - 1 downto 0);
begin
    process (clk, reset) is
    begin
        if reset = '1' then
            i   <= 0;
            j   <= 0;
            cur <= 0;
            bst <= 0;
            okay <= x"ff";
            input <= lfsr_o(7 downto 0);
        elsif rising_edge(clk) then
            if j = 128 then
                if i = 15 then
                    if cur >= bst then
                        bst <= cur;
                    end if;

                    i <= 0;
                    cur <= 0;
                    okay <= x"ff";
                else
                    i <= i + 1;
                end if;

                j <= 0;
                input <= lfsr_o(7 downto 0);
            else
                if j >= 16 then
                    case func is
                        when FIT_1 =>
                            cur <= cur + count_ones(input xor (not ca_out));
                        when FIT_2 =>
                            if input = ca_out then
                                cur <= cur + 1;
                            else
                                cur <= cur;
                            end if;
                        when FIT_3 =>
                            okay <= okay and (input xor (not ca_out));
                            cur <= count_ones(okay);
                    end case;
                end if;
                j <= j + 1;
            end if;
        end if;
    end process;

    best  <= bst;
    good  <= '1' when i = 15 and j = 128 and cur >= bst else '0';
    bad   <= '1' when i = 15 and j = 128 and cur <  bst else '0';
    ca_in <= input;

    LFSR : entity work.lfsr
    generic map (
        reset_value => 12345678
    )
    port map (
        clk => clk,
        o   => lfsr_o
    );
end ea_arch;
