library ieee;
use ieee.std_logic_1164.all;

package fitness_pkg is
    type fitness_function is
        ( FIT_1
        , FIT_2
        , FIT_3
        );

    function decode_fitness_func(v : std_logic_vector(31 downto 0)) return fitness_function;
end fitness_pkg;

package body fitness_pkg is
    function decode_fitness_func(v : std_logic_vector(31 downto 0))
    return fitness_function is
    begin
        case v is
            when x"00000000" => return FIT_1;
            when x"00000001" => return FIT_2;
            when x"00000002" => return FIT_3;
            when others      => return FIT_1;
        end case;
    end function;

end package body;

library ieee;
use ieee.std_logic_1164.all;
use work.fitness_pkg.all;

entity fitness is
    generic (
        width : integer := 8
    );
    port (
        clk   : in  std_logic;
        reset : in  std_logic;
        i     : in  std_logic_vector(width - 1 downto 0);
        o     : in  std_logic_vector(width - 1 downto 0);
        good  : out std_logic;
        bad   : out std_logic;
        best  : out natural;
        func  : in  fitness_function
    );
end fitness;

architecture fitness_arch of fitness is
    signal bst : natural;
    signal cur : natural;
    signal cnt : natural;

    function count_ones(v : std_logic_vector) return natural is
        variable cnt : natural := 0;
    begin
        for i in v'low to v'high loop
            if v(i) = '1' then
                cnt := cnt + 1;
            end if;
        end loop;
        return cnt;
    end function;
begin
    process (clk, reset) is
    begin
        if reset = '1' then
            bst <= 0;
            cur <= 0;
            cnt <= 0;
        elsif rising_edge(clk) then
            if cnt = 128 then
                cnt <= 0;
                cur <= 0;
                if cur >= bst then
                    bst <= cur;
                end if;
            elsif cnt > 16 then
                cnt <= cnt + 1;

                if func = FIT_1 then
                    cur <= cur + count_ones(i xor (not o));
                else
                    if i = o then
                        cur <= cur + 1;
                    else
                        cur <= cur;
                    end if;
                end if;
            else
                cnt <= cnt + 1;
                cur <= cur;
            end if;
        end if;
    end process;

    best <= bst;
    good <= '1' when cnt = 128 and cur >= bst else '0';
    bad  <= '1' when cnt = 128 and cur <  bst else '0';
end fitness_arch;
