-- Linear feedback shift register
-- Can be used as a pseudo random generator.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity lfsr is
    generic (
        reset_value : integer := 0
    );
    port (
        clk   : in  std_logic;
        o     : out std_logic_vector(31 downto 0)
    );
end lfsr;

architecture lfsr_arch of lfsr is
    signal state : std_logic_vector(31 downto 0) := std_logic_vector(to_unsigned(reset_value, 32));
begin
    process (clk) is
    begin
        if rising_edge(clk) then
            state(31 downto 1) <= state(30 downto 0);
            state(0) <= state(31) xnor state(21) xnor state(1) xnor state(0);
        end if;
    end process;
    
    o <= state;
end lfsr_arch;
