library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity small_ram is
    generic (
        addr_width : integer := 8
    );
    port (
        clk   : in std_logic;
        raddr : in  std_logic_vector(addr_width - 1 downto 0);
        rdata : out std_logic_vector(7 downto 0);
        waddr : in  std_logic_vector(addr_width - 1 downto 0);
        wdata : in  std_logic_vector(7 downto 0);
        wen   : in  std_logic
    );
end small_ram;

architecture small_ram_arch of small_ram is
    subtype index is natural range 0 to (2**addr_width - 1);
    type ram_type is array (integer range <>) of std_logic_vector(7 downto 0);

    signal mem : ram_type(0 to 2**addr_width -1);

    signal ridx : index;
    signal widx : index;
begin
    ridx <= to_integer(unsigned(raddr));
    widx <= to_integer(unsigned(waddr));

    process (clk) is
    begin
        if rising_edge(clk) then
            rdata <= mem(ridx);

            if wen = '1' then
                mem(widx) <= wdata;
            end if;
        end if;
    end process;
end small_ram_arch;


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity ram is
    generic (
        addr_width : integer := 8;
        data_width : integer := 64
    );
    port (
        clk   : in  std_logic;
        raddr : in  std_logic_vector(addr_width - 1 downto 0);
        rdata : out std_logic_vector(data_width - 1 downto 0);
        waddr : in  std_logic_vector(addr_width - 1 downto 0);
        wdata : in  std_logic_vector(data_width - 1 downto 0);
        wstrb : in  std_logic_vector(data_width / 8 - 1 downto 0)
    );
end ram;

architecture ram_arch of ram is
begin
    GEN_SR :
    for i in 0 to (data_width / 8 - 1) generate
        SR : entity work.small_ram
        generic map (
            addr_width => addr_width
        )
        port map (
            clk   => clk,
            raddr => raddr,
            rdata => rdata(i * 8 + 7 downto i * 8),
            waddr => waddr,
            wdata => wdata(i * 8 + 7 downto i * 8),
            wen   => wstrb(i)
        );
    end generate;
end ram_arch;
