library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.axi.all;
use work.matrix_pkg.all;
use work.sys_regs_pkg.all;
use work.fitness_pkg.all;

library unisim;
use unisim.vcomponents.all;

entity toplevel is
    port (
        xu              : inout std_logic_vector(143 downto 0);
        xl              : inout std_logic_vector(143 downto 0);

        yu              : inout std_logic_vector(143 downto 0);

        m_awid          : in  std_logic_vector(1 downto 0);
        m_awaddr        : in  std_logic_vector(15 downto 0);
        m_awlen         : in  std_logic_vector(1 downto 0);
        m_awsize        : in  std_logic;
        m_awburst       : in  std_logic;
        m_awlock        : in  std_logic;
        m_awcache       : in  std_logic_vector(1 downto 0);
        m_awprot        : in  std_logic_vector(1 downto 0);
        m_awvalid       : in  std_logic;
        awready         : out std_logic;

        m_wid           : in  std_logic_vector(1 downto 0);
        m_wdata         : in  std_logic_vector(31 downto 0);
        m_wstrb         : in  std_logic_vector(3 downto 0);
        m_wlast         : in  std_logic;
        m_wvalid        : in  std_logic;
        wready          : out std_logic;

        m_bid           : out std_logic_vector(2 downto 0);
        m_bresp         : out std_logic;
        m_bvalid        : out std_logic;
        bready          : in  std_logic;

        m_arid          : in  std_logic_vector(1 downto 0);
        m_araddr        : in  std_logic_vector(15 downto 0);
        m_arlen         : in  std_logic_vector(1 downto 0);
        m_arsize        : in  std_logic;
        m_arburst       : in  std_logic;
        m_arlock        : in  std_logic;
        m_arcache       : in  std_logic_vector(1 downto 0);
        m_arprot        : in  std_logic_vector(1 downto 0);
        m_arvalid       : in  std_logic;
        arready         : out std_logic;

        m_rid           : out std_logic_vector(1 downto 0);
        m_rdata         : out std_logic_vector(31 downto 0);
        m_rresp         : out std_logic;
        m_rlast         : out std_logic;
        m_rvalid        : out std_logic;
        rready          : in  std_logic;

        z               : inout std_logic_vector(127 downto 0);
        zu              : inout std_logic_vector(234 downto 128);
        zl              : inout std_logic_vector(233 downto 200);

--        ram_a_sd        : inout std_logic_vector(63 downto 0);
--        ram_a_sa        : out   std_logic_vector(23 downto 0);
--        ram_a_snce0     : out   std_logic;
--        ram_a_snce1     : out   std_logic;
--        ram_a_sclk0     : out   std_logic;
--        ram_a_sclk1     : out   std_logic;
--        ram_a_snwbyte   : out   std_logic_vector(7 downto 0);
--        ram_a_snoe      : out   std_logic;
--        ram_a_sadvnld   : out   std_logic;
--        ram_a_snwe      : out   std_logic;
--        ram_a_sncke     : out   std_logic;
--        ram_a_smode     : out   std_logic;
--
--        ram_b_sd        : inout std_logic_vector(63 downto 0);
--        ram_b_sa        : out   std_logic_vector(23 downto 0);
--        ram_b_snce0     : out   std_logic;
--        ram_b_snce1     : out   std_logic;
--        ram_b_sclk0     : out   std_logic;
--        ram_b_sclk1     : out   std_logic;
--        ram_b_snwbyte   : out   std_logic_vector(7 downto 0);
--        ram_b_snoe      : out   std_logic;
--        ram_b_sadvnld   : out   std_logic;
--        ram_b_snwe      : out   std_logic;
--        ram_b_sncke     : out   std_logic;
--        ram_b_smode     : out   std_logic;

        fnce            : out   std_logic;

        nxl_fold        : out   std_logic;
        nxu_fold        : out   std_logic;
        nx_thru         : out   std_logic;

        nyl_fold        : out   std_logic;
        nyu_fold        : out   std_logic;
        ny_thru         : out   std_logic;

        nzl_fold        : out   std_logic;
        nzu_fold        : out   std_logic;
        nz_thru         : out   std_logic;

        user_sw         : in    std_logic_vector(7 downto 0);
        user_led        : out   std_logic_vector(7 downto 0);
        npb             : in    std_logic;

        nrtcken         : out   std_logic;

        clk_glbl_nen    : out   std_logic;

        d_tck, d_tdi    : in    std_logic;
        d_tdo, d_rtck   : out   std_logic;

        nsyspor         : in    std_logic;
        nsysrst         : in    std_logic;

        clk_neg_dn_out  : out   std_logic;
        clk_neg_up_out  : out   std_logic;
        clk_pos_dn_out  : out   std_logic;
        clk_pos_up_out  : out   std_logic;

        clk_neg_dn_in   : in    std_logic;
        clk_neg_up_in   : in    std_logic;
        clk_pos_dn_in   : in    std_logic;
        clk_pos_up_in   : in    std_logic;

        clk0            : in    std_logic;
        clk1            : in    std_logic;
        clk2            : in    std_logic;
        clk_extern      : in    std_logic;
        clk_buf_loop    : in    std_logic;
        clk_24mhz_fpga  : in    std_logic;
        clk_global_out  : out   std_logic;
        clk_in_minus1   : in    std_logic;
        clk_loop0_in    : in    std_logic;
        clk_loop1_out   : out   std_logic;
        clk_loop0_out   : out   std_logic;
        clk_loop1_in    : in    std_logic;
        clk_out_to_buf  : out   std_logic);
end toplevel;

architecture arch_top of toplevel is
    signal clk : std_logic;

    -- Demultiplexed signals
    signal aw    : axi_addr_channel;
    signal ar    : axi_addr_channel;
    signal w     : axi_write_data_channel;

    signal sl_r  : axi_read_data_channel;
    signal sl_rready : std_logic;

    signal sl_b  : axi_write_resp_channel;
    signal sl_bready : std_logic;

    signal ma_r  : axi_read_data_channel;
    signal ma_rready : std_logic;

    signal ma_b  : axi_write_resp_channel;
    signal ma_bready : std_logic;

    -- Non-multiplexed signals
    signal aclk     : std_logic;
    signal aresetn  : std_logic;
    signal reset    : std_logic;

    signal dcm_out : std_logic;

    -- Matrix signals
    constant slave_count : integer := 1;
    signal matrix_master_in  : matrix_master_in;
    signal matrix_master_out : matrix_out;
    signal matrix_slaves_out : matrix_slave_in_array(0 to slave_count - 1);
    signal matrix_slaves_in  : matrix_out_array(0 to slave_count - 1);

    signal ca_func   : fitness_function;
    signal ca_best   : natural;
    signal ca_input  : std_logic_vector(31 downto 0);
    signal ca_output : std_logic_vector(7 downto 0);
    signal ca_in     : std_logic_vector(7 downto 0);
    signal ca_use_sw : boolean;
    signal ca_evolve : std_logic;
    signal ca_reset  : std_logic;
begin
	reset <= not nsysrst;

	DCM_ENT : dcm_base
	generic map (
		dcm_performance_mode => "MAX_RANGE"
	)
	port map (
		clkin => clk_in_minus1,
		clkfb => clk,
		clk0  => dcm_out,
		rst   => reset
	);

	BUFG_ENT : bufg
	port map (
		i => dcm_out,
		o => clk
	);

    aclk      <= clk;
    aresetn   <= nsysrst;
    ma_rready <= rready;
    ma_bready <= bready;

    -- Setup fold/thru
    nxl_fold <= '1';
    nxu_fold <= '1';
    nx_thru  <= '0';

    nyl_fold <= '1';
    nyu_fold <= '1';
    ny_thru  <= '0';

    nzl_fold <= '1';
    nzu_fold <= '1';
    nz_thru  <= '1';

    -- Disable RTCK
    nrtcken  <= '0';

    -- Disable flash
    fnce <= '1';

    -- Disable global clock
    clk_glbl_nen <= '1';

    -- Various clock stuff..
    clk_loop0_out  <= '0';
    clk_loop1_out  <= '0';
    clk_global_out <= '0';
    clk_out_to_buf <= '0';
    clk_neg_dn_out <= '0';
    clk_neg_up_out <= '0';
    clk_pos_dn_out <= '0';
    clk_pos_up_out <= '0';

    -- Route through JTAG signals
    d_tdo  <= d_tdi;
    d_rtck <= d_tck;

	MATRIX : entity work.matrix
	generic map (
		slave_count => slave_count
	)
	port map (
        clk        => clk,
        reset      => reset,
        master_in  => matrix_master_in,
        master_out => matrix_master_out,
        slaves_in  => matrix_slaves_in,
        slaves_out => matrix_slaves_out
	);

    SR : entity work.sys_regs
    port map (
        clk        => clk,
        reset      => reset,
        matrix_out => matrix_slaves_in(0),
        matrix_in  => matrix_slaves_out(0),
        ca_best    => ca_best,
        ca_func    => ca_func,
        ca_input   => ca_input,
        ca_output  => ca_output,
        ca_use_sw  => ca_use_sw,
        ca_evolve  => ca_evolve,
        ca_reset   => ca_reset
    );

	-- AXI entities
    AXI_DEMUX : entity work.axi_demux
    port map (
        clk       => clk,
        m_awid    => m_awid   ,
        m_awaddr  => m_awaddr ,
        m_awlen   => m_awlen  ,
        m_awsize  => m_awsize ,
        m_awburst => m_awburst,
        m_awlock  => m_awlock ,
        m_awcache => m_awcache,
        m_awprot  => m_awprot ,
        m_awvalid => m_awvalid,

        m_wid     => m_wid    ,
        m_wdata   => m_wdata  ,
        m_wstrb   => m_wstrb  ,
        m_wlast   => m_wlast  ,
        m_wvalid  => m_wvalid ,

        m_bid     => m_bid    ,
        m_bresp   => m_bresp  ,
        m_bvalid  => m_bvalid ,

        m_arid    => m_arid   ,
        m_araddr  => m_araddr ,
        m_arlen   => m_arlen  ,
        m_arsize  => m_arsize ,
        m_arburst => m_arburst,
        m_arlock  => m_arlock ,
        m_arcache => m_arcache,
        m_arprot  => m_arprot ,
        m_arvalid => m_arvalid,

        m_rid     => m_rid    ,
        m_rdata   => m_rdata  ,
        m_rresp   => m_rresp  ,
        m_rlast   => m_rlast  ,
        m_rvalid  => m_rvalid ,

        aw        => aw,
        ar        => ar,
        w         => w,
        r         => ma_r,
        b         => ma_b
    );

    AXI_SLAVE : entity work.axi_slave
    port map (
        aclk    => aclk   ,
        aresetn => aresetn,

        aw      => aw,
        awready => awready,
        ar      => ar,
        arready => arready,
        w       => w,
        wready  => wready ,
        b       => sl_b,
        bready  => sl_bready ,
        r       => sl_r,
        rready  => sl_rready,

        matrix_out => matrix_master_out,
        matrix_in  => matrix_master_in
    );

    REG_SLICE : entity work.axi_regslice
    port map (
        clk      => aclk,
        nreset   => aresetn,

        s_b      => sl_b,
        s_bready => sl_bready,
        m_b      => ma_b,
        m_bready => ma_bready,
        s_r      => sl_r,
        s_rready => sl_rready,
        m_r      => ma_r,
        m_rready => ma_rready
    );

    ca_in <= user_sw when ca_use_sw else ca_input(7 downto 0);
    CA : entity work.ca
    generic map (
        width  => 8,
        height => 8
    )
    port map (
        clk   => clk,
        reset => ca_reset,
        i     => ca_in,
        o     => ca_output,
        best  => ca_best,
        func  => ca_func,
        evolve => ca_evolve
    );

    user_led <= ca_output;
end arch_top;
