#define _XOPEN_SOURCE
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "ea.h"
#include "ca.h"

#define GENS 10000
#define CA_W 8
#define CA_H 8


#if 1
//#define FIT_FUNC(o, i) (count_bits((o) ^ (~(i) & 0xff)))
#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))
#define FIT_MAX 8
#else
//#define FIT_FUNC(o, i) ((o) == (~(i) & 0xff))
#define FIT_FUNC(o, i) ((o) == ((i) & 0xff))
#define FIT_MAX 1
#endif

#include "util.h"

static ca_t *ca;

static void
check_ca(uni_genome_t *gen)
{
	lut_t luts[CA_W * CA_H];

	for (int i = 0; i < CA_W * CA_H; i++) {
		luts[i] = gen->lut;
	}

	ca_set_luts(ca, luts);

	for (uint32_t in = 0; in < 256; in++) {
		uint32_t *outs = ca_run(ca, in, 128, NULL);

		for (int j = 16; j < 128; j++) {
			if (FIT_FUNC(outs[j], in) != FIT_MAX) {
				printf("\t[%i] %08x != %08x\n", j, in, outs[j]);
				break;
			}
		}

		free(outs);
	}
}

static void *
run_n(ea_cfg_t *ea_cfg, const char *name, int n, int gens)
{
	ea_result_t *results;
	uni_genome_t *g;

	results = ea_one_plus_n(ea_cfg, n, gens);

	printf("# %s\n", name);
	for (int i = 0; i < gens; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
	}
	printf("\n\n");
	/*
	fprintf(stderr, "1+%i: %08x\n",
	        n,
	        ((uni_genome_t*)results[gens - 1].gen_best)->lut);
	        */

//	free(results);
	return results[gens - 1].gen_best;
}

static void
run_pop(ea_cfg_t *ea_cfg, int pop_size)
{
	ea_result_t *results;
	uni_genome_t *g;

	results = ea_ga(ea_cfg, pop_size, GENS);

	printf("# GA %i\n", pop_size);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
	}
	/*
	printf("\n\n");
	fprintf(stderr, "GA %i: %08x\n",
	        pop_size,
	        ((uni_genome_t*)results[GENS - 1].gen_best)->lut);
	        */

	free(results);
}

static uint32_t lut_best;

static void *
both_new(void)
{
	nu_genome_t *ret = malloc(sizeof(nu_genome_t));

	for (int y = 0; y < CA_H; y++) {
		for (int x = 0; x < CA_W; x++) {
			ret->luts[x + y * CA_W] = lut_best;
		}
	}

	return ret;
}

int
main(int argc, char *argv[])
{
	ca = ca_new(CA_W, CA_H);
	uni_cfg.user_data = ca;
	nu_cfg.user_data = ca;
	nu_cfg.new = both_new;

	srand48(time(NULL));

	printf("# Max\n0 %lf\n%i %lf\n\n\n", 1792.0 * FIT_MAX, GENS - 1, 1792.0 * FIT_MAX);

	lut_best = ((uni_genome_t*)run_n(&uni_cfg, "uni", 4, GENS))->lut;

//	run_pop(&nu_cfg, 16);
	run_n(&nu_cfg, "nu", 4, GENS);


	return 0;
}
