#include "ea.h"
#include <stdlib.h>


ea_result_t *
ea_one_plus_n(const ea_cfg_t *cfg, int n, int generations)
{
	return ea_one_plus_n_max(cfg, n, generations, 1e100);
}
ea_result_t *
ea_one_plus_n_max(const ea_cfg_t *cfg, int n, int generations, double max)
{
	ea_result_t *res = malloc(sizeof(ea_result_t) * generations);

	int gen;
	double fit_best = -1e100;
	void *gen_best = cfg->new();

	for (gen = 0; gen < generations && max > fit_best; gen++) {
		void *gen_old = cfg->copy(gen_best);
		double tot;

		tot = fit_best = cfg->calc_fitness(gen_best, cfg->user_data);

		for (int i = 0; i < n; i++) {
			void *gen_new = cfg->mutate(gen_old);
			double fit_new = cfg->calc_fitness(gen_new, cfg->user_data);

			tot += fit_new;

			if (fit_new >= fit_best) {
				free(gen_best);

				fit_best = fit_new;
				gen_best = gen_new;
			} else {
				free(gen_new);
			}
		}

		free(gen_old);

		res[gen].fit_best = fit_best;
		res[gen].fit_avg = tot / (1 + n);
		res[gen].gen_best = cfg->copy(gen_best);
	}

	free(gen_best);

	for (int i = gen; i < generations; i++) {
		res[i].fit_best = res[gen - 1].fit_best;
		res[i].fit_avg  = res[gen - 1].fit_avg;
		res[i].gen_best = cfg->copy(res[gen - 1].gen_best);
	}

	return res;
}

typedef struct {
	void *gen;
	double fit;
} pop_t;

static int
pop_cmp(const void *a, const void *b)
{
	const pop_t *pa = a;
	const pop_t *pb = b;

	if (pa->fit < pb->fit) {
		return -1;
	} else if (pa->fit > pb->fit) {
		return 1;
	} else {
		return 0;
	}
}

static void*
pop_sel(const pop_t *pop, int pop_size)
{
	void *ret = NULL;
	double best = -1e100;

	for (int k = 0; k < 2; k++) {
		int idx = rand() % pop_size;

		if (pop[idx].fit > best) {
			ret = pop[idx].gen;
			best = pop[idx].fit;
		}
	}

	return ret;
}

ea_result_t *
ea_ga(const ea_cfg_t *cfg, int pop_size, int generations)
{
	ea_result_t *res = malloc(sizeof(ea_result_t) * generations);
	pop_t pop[pop_size];

	for (int i = 0; i < pop_size; i++) {
		pop[i].gen = cfg->new();
	}

	for (int gen = 0; gen < generations; gen++) {
		void *new_pop[pop_size];

		for (int i = 0; i < pop_size; i++) {
			pop[i].fit = cfg->calc_fitness(pop[i].gen, cfg->user_data);
		}

		if (gen == generations - 1) {
			qsort(pop, pop_size, sizeof(pop_t), pop_cmp);
			res->last_gen = malloc(sizeof(void*) * pop_size);
			res->last_fit = malloc(sizeof(double) * pop_size);

			for (int i = 0; i < pop_size; i++) {
				res->last_gen[i] = cfg->copy(pop[i].gen);
				res->last_fit[i] = pop[i].fit;
			}
		}

		for (int i = 0; i < pop_size; i += 2) {
			void *a, *b;

			a = cfg->copy(pop_sel(pop, pop_size));
			b = cfg->copy(pop_sel(pop, pop_size));

			if (rand() < (cfg->cross_rate * RAND_MAX)) {
				void *na, *nb;

				cfg->crossover(a, b, &na, &nb);

				free(a);
				free(b);

				a = na;
				b = nb;
			}

			if (rand() < (cfg->mut_rate * RAND_MAX)) {
				void *na, *nb;

				na = cfg->mutate(a);
				nb = cfg->mutate(b);

				free(a);
				free(b);

				a = na;
				b = nb;
			}

			new_pop[i] = a;
			new_pop[i + 1] = b;
		}
		res[gen].fit_avg = 0;
		res[gen].fit_best = pop[0].fit;
		res[gen].gen_best = pop[0].gen;

		double dist_avg = 0;
		double dist_var = 0;
		int n = cfg->distance == NULL ? 1 : 0;

		for (int i = 0; i < pop_size; i++) {
			for (int j = i + 1; cfg->distance != NULL && j < pop_size; j++) {
				double d = cfg->distance(pop[i].gen, pop[j].gen);
				dist_avg += d;
				dist_var += d * d;
				n++;
			}

			if (pop[i].fit > res[gen].fit_best) {
				free(res[gen].gen_best);
				res[gen].gen_best = pop[i].gen;
				res[gen].fit_best = pop[i].fit;
			} else if (i > 0) {
				free(pop[i].gen);
			}

			pop[i].gen = new_pop[i];
			res[gen].fit_avg += pop[i].fit;
		}

		dist_avg /= n;
		dist_var /= n;
		dist_var -= dist_avg * dist_avg;

		res[gen].fit_avg /= pop_size;
		res[gen].dist_avg = dist_avg;
		res[gen].dist_var = dist_var;
	}

	return res;
}
