#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ca.h"

#include "ea.h"
#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))

#define CA_W 8
#define CA_H 8
#define WIN_SIZE 16

#include "util.h"

#define ITS CA_W
#define STEP 1

// static int goal_compact[] = { 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55 };

int
main(int argc, char *argv[])
{
	ls_genome_t *gen;
	int w, h;

	gen = ls_read(stdin, &w, &h);
	ls_dump(stdout, gen, w, h);

	ca_t *ca;
    uint32_t *outs;
    uint8_t *states;
    int mode = 0;

	if (argc > 1) {
		mode = atoi(argv[1]);
	}

	ca_fft_init();
	fft_init(WIN_SIZE);

	FILE *graph = NULL;
	if (argc > 2) {
		graph = fopen(argv[2], "w");
	}

    ca = ca_new(w, h);
    ls_set_luts(ca, gen, w, h);

	outs = ca_run_initial(ca, 0, WIN_SIZE * 2, &states, gen->initial);
	if (graph != NULL) {
		ca_dump(graph, 0, states, w, h, WIN_SIZE * 2);
	}
	free(outs);
	free(states);
	outs = ca_run_initial(ca, 1, WIN_SIZE * 2, &states, gen->initial);
	if (graph != NULL) {
		ca_dump(graph, 1, states, w, h, WIN_SIZE * 2);
	}
	free(outs);
	free(states);

	fprintf(graph, "\n\n# Data\n");

	for (int i = 0; i < 256; i ++) {
		double fft_in[WIN_SIZE];
		double *spec;
		int even = count_bits(i) % 2 == 0;

		outs = ca_run_initial(ca, i, WIN_SIZE * 2, NULL, gen->initial);

		if (mode) {
			if (graph != NULL) {
				fprintf(graph, "%i %i %i\n", i, even, outs[WIN_SIZE * 2 - 1]);
			}
		} else {
			for (int i = 0; i < WIN_SIZE; i++) {
				fft_in[i] = count_bits(outs[i + WIN_SIZE]);
			}

			spec = fft_run(fft_in);

			if (graph != NULL) {
				fprintf(graph, "%i %i %f\n", i, even, spec_avg(spec, WIN_SIZE / 2 + 1));
			}
		}

		free(outs);
	}

	if (graph != NULL) {
		int array[w * h];

		ls_run(gen, array, w, h);

		fprintf(graph, "\n\n# LS-img\n");
		for (int y = 0; y < h; y++) {
			for (int x = 0; x < w; x++) {
				fprintf(graph, "%i %i %i\n", x, y, array[x + y * w]);
			}
		}

		fclose(graph);
	}

    return 0;
}
