#define _XOPEN_SOURCE
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "ea.h"
#include "ca.h"

#define GENS 10000
#define CA_W 8
#define CA_H 8

#define WIN_SIZE 16


#if 1
//#define FIT_FUNC(o, i) (count_bits((o) ^ (~(i) & 0xff)))
#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))
#define FIT_MAX 8
#else
//#define FIT_FUNC(o, i) ((o) == (~(i) & 0xff))
#define FIT_FUNC(o, i) ((o) == ((i) & 0xff))
#define FIT_MAX 1
#endif

#include "util.h"

static ca_t *ca;

static void
run_n(ea_cfg_t *ea_cfg, int n)
{
	ea_result_t *results;

	results = ea_one_plus_n(ea_cfg, n, GENS);

	printf("# 1+%i\n", n);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
	}
	printf("\n\n");

	free(results);
}

static void
run_pop(ea_cfg_t *ea_cfg, int pop_size)
{
	ea_result_t *results;

	results = ea_ga(ea_cfg, pop_size, GENS);

	printf("# GA %i\n", pop_size);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
	}
	printf("\n\n");

	free(results);
}

static double
fft_nu_fitness(void *g, void *ud)
{
	ca_t *ca = ud;
	nu_genome_t *gen = g;
	int tot = 0;
	uint32_t inputs[128];

	ca_set_luts(ca, gen->luts);

	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 128; j++) {
			inputs[j] = ((j / (1 << i)) % 2) ? 0xff : 0x00;
		}

		uint32_t *outs = ca_run_array(ca, inputs, 128, NULL, 0);
		double fft_in[WIN_SIZE];
		double *spec;

		for (int i = 0; i < WIN_SIZE; i++) {
			fft_in[i] = count_bits(outs[i + WIN_SIZE - 1]);
		}

		for (int j = WIN_SIZE; j < 128 - WIN_SIZE; j++) {
			memmove(fft_in, fft_in + 1, (WIN_SIZE - 1) * sizeof(double));
			fft_in[WIN_SIZE - 1] = count_bits(outs[j + WIN_SIZE - 1]);

			spec = fft_run(fft_in);

//			if (spec_max(spec + 1, WIN_SIZE / 2) == (1 << (3 - i)) - 1) {
			if (spec_max(spec + 1, WIN_SIZE / 2) == (1 << (i)) - 1) {
				tot++;
			}
		}

		free(outs);
	}

	return tot;
}

int
main(int argc, char *argv[])
{
	ca = ca_new(CA_W, CA_H);
	nu_cfg.user_data = ca;
	nu_cfg.calc_fitness = fft_nu_fitness;

	fft_init(WIN_SIZE);

	srand48(time(NULL));

	printf("# Max\n0 %lf\n%i %lf\n\n\n", 384.0 * 1, GENS - 1, 384.0 * 1);

	run_n(&nu_cfg, 1);
	run_n(&nu_cfg, 4);
	run_pop(&nu_cfg, 16);

	return 0;
}
