#define _XOPEN_SOURCE
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "ea.h"
#include "ca.h"

#define GENS 100000
#define CA_W 8
#define CA_H 8

#define WIN_SIZE 16


#if 1
//#define FIT_FUNC(o, i) (count_bits((o) ^ (~(i) & 0xff)))
#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))
#define FIT_MAX 8
#else
//#define FIT_FUNC(o, i) ((o) == (~(i) & 0xff))
#define FIT_FUNC(o, i) ((o) == ((i) & 0xff))
#define FIT_MAX 1
#endif

#include "util.h"

static ca_t *ca;

static void
check_ca(uni_genome_t *gen)
{
	lut_t luts[CA_W * CA_H];

	for (int i = 0; i < CA_W * CA_H; i++) {
		luts[i] = gen->lut;
	}

	ca_set_luts(ca, luts);

	for (uint32_t in = 0; in < 256; in++) {
		uint32_t *outs = ca_run(ca, in, 128, NULL);

		for (int j = 16; j < 128; j++) {
			if (FIT_FUNC(outs[j], in) != FIT_MAX) {
				printf("\t[%i] %08x != %08x\n", j, in, outs[j]);
				break;
			}
		}

		free(outs);
	}
}

static void
run_n(ea_cfg_t *ea_cfg, int n)
{
	ea_result_t *results;
	uni_genome_t *g;

	results = ea_one_plus_n(ea_cfg, n, GENS);

	printf("# 1+%i\n", n);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
	}
	printf("\n\n");
	fprintf(stderr, "1+%i: %08x\n",
	        n,
	        ((uni_genome_t*)results[GENS - 1].gen_best)->lut);

	free(results);
}

static void
run_pop(ea_cfg_t *ea_cfg, int pop_size)
{
	ea_result_t *results;
	uni_genome_t *g;

	results = ea_ga(ea_cfg, pop_size, GENS);

	printf("# GA %i\n", pop_size);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
	}
	printf("\n\n");
	fprintf(stderr, "GA %i: %08x\n",
	        pop_size,
	        ((uni_genome_t*)results[GENS - 1].gen_best)->lut);

	free(results);
}

static double
fft_uni_fitness(void *g, void *ud)
{
	ca_t *ca = ud;
	uni_genome_t *gen = g;
	int tot = 0;
	lut_t luts[CA_W * CA_H];
	uint32_t inputs[128];

	for (int i = 0; i < CA_W * CA_H; i++) {
		luts[i] = gen->lut;
	}

	ca_set_luts(ca, luts);

	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 128; j++) {
			inputs[j] = ((j / (1 << i)) % 2) ? 0xff : 0x00;
		}

		uint32_t *outs = ca_run_array(ca, inputs, 128, NULL, 0);
		double fft_in[WIN_SIZE];
		double *spec;

		for (int i = 0; i < WIN_SIZE; i++) {
			fft_in[i] = count_bits(outs[i + WIN_SIZE - 1]);
		}

		for (int j = WIN_SIZE; j < 128 - WIN_SIZE; j++) {
			memmove(fft_in, fft_in + 1, (WIN_SIZE - 1) * sizeof(double));
			fft_in[WIN_SIZE - 1] = count_bits(outs[j + WIN_SIZE - 1]);

			spec = fft_run(fft_in);

//			if (spec_max(spec + 1, WIN_SIZE / 2) == (1 << (3 - i)) - 1) {
			if (spec_max(spec + 1, WIN_SIZE / 2) == (1 << (i)) - 1) {
				tot++;
			}

			/*
			printf("%i.%i\n", i, j);
			for (int k = 0; k < WIN_SIZE; k++) {
				printf("\t%i => %f\n", k, fft_in[k]);
			}
			printf("\t----\n");

			for (int k = 0; k < WIN_SIZE / 2 + 1; k++) {
				printf("\t%i => %f\n", k, spec[k]);
			}
			*/
		}

		free(outs);
	}

	return tot;
}

int
main(int argc, char *argv[])
{
	ca = ca_new(CA_W, CA_H);
	uni_cfg.user_data = ca;
	uni_cfg.calc_fitness = fft_uni_fitness;

	fft_init(WIN_SIZE);

	srand48(time(NULL));

	printf("# Max\n0 %lf\n%i %lf\n\n\n", 384.0 * 1, GENS - 1, 384.0 * 1);

	run_n(&uni_cfg, 1);
	run_n(&uni_cfg, 4);
	run_pop(&uni_cfg, 16);

	return 0;
}
