#define _XOPEN_SOURCE
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "ea.h"
#include "ca.h"

#define GENS 50000
#define CA_W 8
#define CA_H 8

#define WIN_SIZE 16


#if 1
//#define FIT_FUNC(o, i) (count_bits((o) ^ (~(i) & 0xff)))
#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))
#define FIT_MAX 8
#else
//#define FIT_FUNC(o, i) ((o) == (~(i) & 0xff))
#define FIT_FUNC(o, i) ((o) == ((i) & 0xff))
#define FIT_MAX 1
#endif

#include "util.h"

#define CELL_FROM 1
#define CELL_TO 4
#define TIMES 10

static ca_t *ca;

static void
check_ca(uni_genome_t *gen)
{
	lut_t luts[CA_W * CA_H];

	for (int i = 0; i < CA_W * CA_H; i++) {
		luts[i] = gen->lut;
	}

	ca_set_luts(ca, luts);

	for (uint32_t in = 0; in < 256; in++) {
		uint32_t *outs = ca_run(ca, in, 128, NULL);

		for (int j = 16; j < 128; j++) {
			if (FIT_FUNC(outs[j], in) != FIT_MAX) {
				printf("\t[%i] %08x != %08x\n", j, in, outs[j]);
				break;
			}
		}

		free(outs);
	}
}

static void
run_n(ea_cfg_t *ea_cfg, int n)
{
	ea_result_t *results;
	uni_genome_t *g;

	results = ea_one_plus_n(ea_cfg, n, GENS);

	printf("# 1+%i-%i\n", n, cell_count);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
//		fprintf(stderr, "%08x %lf\n", ((su_genome_t*)results[i].gen_best)->luts[0], results[i].fit_best);
	}
//	printf("\n\n");
	fprintf(stderr, "%f ", results[GENS - 1].fit_best);
	su_dump(stderr, (su_genome_t*)results[GENS - 1].gen_best);

	free(results);
}

static void
run_pop(ea_cfg_t *ea_cfg, int pop_size)
{
	ea_result_t *results;
	uni_genome_t *g;

	results = ea_ga(ea_cfg, pop_size, GENS);

	printf("# GA %i\n", pop_size);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
	}
	printf("\n\n");
	su_dump(stderr, (su_genome_t*)results[GENS - 1].gen_best);

	free(results);
}

static double
fft_su_fitness(void *g, void *ud)
{
	ca_t *ca = ud;
	su_genome_t *gen = g;
	lut_t luts[CA_W * CA_H];

	for (int i = 0; i < CA_W * CA_H; i++) {
		luts[i] = gen->luts[i % cell_count];
	}

	ca_set_luts(ca, luts);

	return fft_fitness1(ca, gen->initial);
}

int
main(int argc, char *argv[])
{
	ca = ca_new(CA_W, CA_H);
	su_cfg.user_data = ca;
	su_cfg.calc_fitness = fft_su_fitness;

	fft_init(WIN_SIZE);

	srand48(time(NULL));

// 	printf("# Max\n0 %lf\n%i %lf\n\n\n", 384.0 * 1, GENS - 1, 384.0 * 1);

//	fprintf(stderr, "%i %i %i\n", 100, 1, GENS);
	for (int i = CELL_FROM; i <= CELL_TO; i++) {
		cell_count = i;
		printf("# %i\n", i);
		for (int j = 0; j < TIMES; j++) {
			run_n(&su_cfg, 4);
			printf("\n");
		}
		printf("\n");
	}

	return 0;
}
