#define _XOPEN_SOURCE
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "ea.h"
#include "ca.h"

#define GENS 10000
#define CA_W 8
#define CA_H 8
#define WIN_SIZE 16

#if 1
//#define FIT_FUNC(o, i) (count_bits((o) ^ (~(i) & 0xff)))
#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))
#define FIT_MAX 8
#else
//#define FIT_FUNC(o, i) ((o) == (~(i) & 0xff))
#define FIT_FUNC(o, i) ((o) == ((i) & 0xff))
#define FIT_MAX 1
#endif

#include "util.h"


static ca_t *ca;

static void
check_ca(nu_genome_t *gen)
{
	ca_set_luts(ca, gen->luts);

	for (uint32_t in = 0; in < 256; in++) {
		uint32_t *outs = ca_run(ca, in, 128, NULL);

		for (int j = 16; j < 128; j++) {
			if (FIT_FUNC(outs[j], in) != FIT_MAX) {
				printf("\t[%i] %08x != %08x\n", j, in, outs[j]);
				break;
			}
		}

		free(outs);
	}
}

static nu_genome_t *wide_best = NULL;

static void
run_n(ea_cfg_t *ea_cfg, int w, int n, double max)
{
	ea_result_t *results;
	nu_genome_t *g;

	results = ea_one_plus_n_max(ea_cfg, n, GENS, max);

	printf("# %i\n", w);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
		/*
		g = results[GENS - i].gen_best;
		printf("\t%.0lf\n", results[GENS - i].fit_best);//, g->lut);

		if (results[GENS - i].fit_best >= 1792 * FIT_MAX) {
			check_ca(g);
		}
		*/
	}
	printf("\n\n");

	wide_best = nu_copy(results[GENS - 1].gen_best);

	char filename[128];
	sprintf(filename, "gen%i.nu", w);

	nu_dump(filename, wide_best, w);
	free(results);
}

static void
run_pop(ea_cfg_t *ea_cfg, int pop_size)
{
	ea_result_t *results;
	nu_genome_t *g;

	results = ea_ga(ea_cfg, pop_size, GENS);

	printf("# GA %i:\n", pop_size);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
		/*
		g = results[GENS - i].gen_best;
		printf("\t%.0lf\n", results[GENS - i].fit_best);//, g->lut);

		if (results[GENS - i].fit_best >= 1792 * FIT_MAX) {
			check_ca(g);
		}
		*/
	}
	printf("\n\n");

	free(results);
}

static int width = 0;


static int
spec_max_(const double *spec, int len)
{
	int idx = 0;
	int max = spec[0];

	for (int i = 0; i < len; i++) {
		if (max < spec[i]) {
			idx = i;
			max = spec[i];
		}
	}

	return idx;
}
static double
fft_wide_fitness(void *g, void *ud)
{
	ca_t *ca = ud;
	nu_genome_t *gen = g;
	int tot = 0;
	uint32_t inputs[128];
	uint32_t luts[CA_H * width];

	for (int x = 0; x < width; x++) {
		for (int y = 0; y < CA_H; y++) {
			luts[x + y * width] = gen->luts[x + y * CA_W];
		}
	}

	ca_set_luts(ca, luts);

	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < 128; j++) {
			inputs[j] = ((j / (1 << i)) % 2) ? 0xff : 0x00;
		}

		uint32_t *outs = ca_run_array(ca, inputs, 128, NULL, 0);
		double fft_in[WIN_SIZE];
		double *spec;

		for (int i = 0; i < WIN_SIZE; i++) {
			fft_in[i] = count_bits(outs[i + WIN_SIZE - 1]);
		}

		for (int j = WIN_SIZE; j < 128 - WIN_SIZE; j++) {
			memmove(fft_in, fft_in + 1, (WIN_SIZE - 1) * sizeof(double));
			fft_in[WIN_SIZE - 1] = count_bits(outs[j + WIN_SIZE - 1]);

			spec = fft_run(fft_in);

			if (spec_max(spec + 1, WIN_SIZE / 2) == (1 << (3 - i)) - 1) {
//			if (spec_max(spec + 1, WIN_SIZE / 2) == (1 << (i)) - 1) {
				tot++;
			}
		}

		free(outs);
	}

	return tot;
}

static void *
wide_new(void)
{
	return wide_best;
}

int
main(int argc, char *argv[])
{
	nu_cfg.calc_fitness = fft_wide_fitness;
	nu_cfg.new = wide_new;

	fft_init(WIN_SIZE);
	wide_best = nu_new();

	srand48(time(NULL));

	printf("# Max\n0 %lf\n%i %lf\n\n\n", 384.0 * 1, GENS - 1, 384.0 * 1);

	for (int w = 1; w <= 8; w++) {
		width = w;
		ca = ca_new(w, CA_H);
		nu_cfg.user_data = ca;
		run_n(&nu_cfg, w, 4, 384);
	}

	return 0;
}
