#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "ca.h"

#define CA_W 8
#define CA_H 8

#define ITS 64

int
main(int argc, char *argv[])
{
	if (argc < 2) {
		fprintf(stderr,
		        "Too few arguments!\n"
		        "Usage: %s <FILE> <OUT>\n",
		        argv[0]);
		exit(1);
	}

	ca_t *ca;
    uint8_t *states;
    uint32_t in[ITS];
    uint32_t *outs;
    uint32_t initial;
	int w;
	lut_t *luts;
	FILE *f = fopen(argv[1], "r");

	ca_fft_init();

	if (f == NULL) {
		fprintf(stderr,
		        "Could not open %s - %s\n", argv[1], strerror(errno));
	}
	fscanf(f, "%i %x", &w, &initial);
	luts = malloc(sizeof(lut_t) * w * CA_H);

	for (int y = 0; y < CA_H; y++) {
		for (int x = 0; x < w; x++) {
			fscanf(f, "%x", &luts[x + y * w]);
			printf("luts[%i, %i] = %08x\n", x, y, luts[x + y * w]);
		}
	}

    ca = ca_new(w, CA_H);
	ca_set_luts(ca, luts);

	FILE *graph = NULL;

	if (argc > 2)
		graph = fopen(argv[2], "w");

	for (int i = 0; i < 4; i++) {
		for (int j = 0; j < ITS; j++) {
			in[j] = (j / (1 << i) % 2) ? 0xff : 0x00;
		}

		outs = ca_run_array(ca, in, ITS, &states, initial);
		if (argc > 2) {
			ca_dump(graph, i + 1, states, w, CA_H, ITS);
		} else {
			ca_anim(states, w, CA_H, ITS);
		}
	}

	if (argc > 2)
		fclose(graph);

    free(outs);
    free(states);
    return 0;
}
