#define _XOPEN_SOURCE
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "ea.h"
#include "ca.h"

#define GENS 10000
#define POP_SIZE 64
#define CA_W 8
#define CA_H 8

#define WIN_SIZE 16
#define STEP 11


#if 1
//#define FIT_FUNC(o, i) (count_bits((o) ^ (~(i) & 0xff)))
#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))
#define FIT_MAX 8
#else
//#define FIT_FUNC(o, i) ((o) == (~(i) & 0xff))
#define FIT_FUNC(o, i) ((o) == ((i) & 0xff))
#define FIT_MAX 1
#endif

#define GEN
#define GA

#include "util.h"

#define CELL_FROM 2
#define CELL_TO 4
#ifdef GA
#define TIMES 1
#else
#define TIMES 10
#endif

static ca_t *ca;

static void
run_n(ea_cfg_t *ea_cfg, int n)
{
	ea_result_t *results;
	uni_genome_t *g;

	results = ea_one_plus_n(ea_cfg, n, GENS);

	printf("# 1+%i-%i\n", n, cell_count);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf %f %f %f %i\n",
		       i,
		       results[i].fit_best, results[i].fit_avg,
		       results[i].dist_avg, results[i].dist_var,
		       ls_size((ls_genome_t*)results[i].gen_best));
//		fprintf(stderr, "%08x %lf\n", ((su_genome_t*)results[i].gen_best)->luts[0], results[i].fit_best);
	}
//	printf("\n\n");
	fprintf(stderr, "\n%f\n", results[GENS - 1].fit_best);
	ls_dump(stderr, (ls_genome_t*)results[GENS - 1].gen_best, CA_W, CA_H);

	for (int i = 0; i < GENS; i++) {
		free(results[i].gen_best);
	}
	free(results);
}

static void
run_pop(ea_cfg_t *ea_cfg, int pop_size)
{
	ea_result_t *results;
	uni_genome_t *g;

	results = ea_ga(ea_cfg, pop_size, GENS);

	printf("# GA %i\n", pop_size);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf %f %f %f %i\n",
		       i,
		       results[i].fit_best, results[i].fit_avg,
		       results[i].dist_avg, results[i].dist_var,
		       ls_size((ls_genome_t*)results[i].gen_best));
	}
//	printf("\n\n");
	fprintf(stderr, "%f ", results[GENS - 1].fit_best);
	ls_dump(stderr, (ls_genome_t*)results[GENS - 1].gen_best, CA_W, CA_H);

	for (int i = 0; i < GENS; i++) {
		free(results[i].gen_best);
	}
	free(results);
}

static int inputs[] =
{
#if CA_H == 7
#define IN_MASK 0x7f
	0,
	1,8,2,64,16,4,32,
	36,33,20,34,48,3,40,72,
	100,74,41,49,52,42,104,38,
	75,39,51,45,58,54,108,99,
	107,59,122,62,55,118,47,31,
	119,126,63,123,111,95,125,
	127,
#elif CA_H == 8
#define IN_MASK 0xff
	/*
	0,
	16,1,4,128,
	17,132,34,20,
	21,81,22,152,
	62,205,185,218,
	187,126,215,222,
	239,223,254,127,
	255
	*/
	/*
	0,
	16,4,2,8,32,64,128,1,
	10,130,144,18,20,36,6,68,
	84,162,52,134,38,49,67,112,
	171,218,182,234,179,155,59,167,
	63,246,231,243,237,252,207,245,
	247,127,223,191,253,254,239,251,
	255,
	*/
	0,
	16,1,64,2,4,8,32,128,
	96,18,3,24,5,6,72,136,68,144,17,129,192,9,36,80,33,66,48,132,160,34,40,65,12,20,10,130,
	146,112,200,69,14,164,133,152,131,50,162,176,41,38,13,137,148,44,26,88,196,49,67,82,21,56,42,134,11,19,208,37,81,76,74,52,7,22,28,138,98,168,161,104,25,73,194,145,70,193,100,97,224,35,84,140,
	118,236,151,244,227,248,109,218,62,55,233,182,171,211,110,230,31,241,91,214,103,107,47,217,229,234,87,143,124,155,242,188,185,93,213,199,94,181,179,186,158,206,117,174,203,205,167,115,122,173,121,157,79,59,61,220,
	175,187,123,250,222,125,111,219,252,183,159,63,246,238,190,231,249,221,243,237,207,215,126,245,119,189,95,235,
	239,223,191,253,251,127,247,254,
	255,
#elif CA_H == 9
#define IN_MASK 0x1ff
	0,
	4,2,32,16,64,256,128,8,
	384,40,257,12,66,20,96,65,
	42,152,416,261,448,73,100,224,
	294,101,46,170,297,390,198,78,
	121,358,143,122,230,482,339,157,
	437,444,311,350,359,190,411,485,
	508,415,251,445,127,319,381,367,
	479,507,383,509,255,495,503,510,
	511,
#endif
};

static double
ls_fitness(void *g, void *ud)
{
	ca_t *ca = ud;
	ls_genome_t *gen = g;
	uint32_t *out;
	int correct = 0;
	int size = sizeof(inputs) / sizeof(int);

	ls_set_luts(ca, gen, CA_W, CA_H);

	for (int i = 0; i < size; i++) {
		out = ca_run_initial(ca, inputs[i], CA_W * 4, NULL, gen->initial);

		uint32_t answer = out[CA_W * 4 - 1];
		int min = count_bits(inputs[i]) < ((CA_H + 1) / 2);

		if ((!min && answer == IN_MASK) || (min && answer == 0x00)) {
			correct++;
		}

		free(out);
	}
	return correct;
}

int
main(int argc, char *argv[])
{
	ca = ca_new(CA_W, CA_H);
	ls_cfg.user_data = ca;
	ls_cfg.calc_fitness = ls_fitness;

	fft_init(WIN_SIZE);

	_srand();

// 	printf("# Max\n0 %lf\n%i %lf\n\n\n", 384.0 * 1, GENS - 1, 384.0 * 1);

//	fprintf(stderr, "%i %i %i\n", 100, 1, GENS);
	for (int i = CELL_FROM; i <= CELL_TO; i++) {
		ls_types = i;
		printf("# %i\n", i);
		for (int j = 0; j < TIMES; j++) {
#ifdef GA
			run_pop(&ls_cfg, POP_SIZE);
#else
			run_n(&ls_cfg, 4);
#endif
			printf("\n");
		}
		printf("\n");
	}

	return 0;
}
