#define _XOPEN_SOURCE
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "ea.h"
#include "ca.h"

#define GENS 10000
#define POP_SIZE 64
#define CA_W 8
#define CA_H 8

#define WIN_SIZE 16


#if 1
//#define FIT_FUNC(o, i) (count_bits((o) ^ (~(i) & 0xff)))
#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))
#define FIT_MAX 8
#else
//#define FIT_FUNC(o, i) ((o) == (~(i) & 0xff))
#define FIT_FUNC(o, i) ((o) == ((i) & 0xff))
#define FIT_MAX 1
#endif

// #define GEN

#include "util.h"

#define CELL_FROM 2
#define CELL_TO 4
#define TIMES 1

static ca_t *ca;

static void
run_n(ea_cfg_t *ea_cfg, int n)
{
	ea_result_t *results;
	uni_genome_t *g;

	results = ea_one_plus_n(ea_cfg, n, GENS);

	printf("# 1+%i-%i\n", n, cell_count);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf %f %i\n", i, results[i].fit_best, results[i].fit_avg, ls_size((ls_genome_t*)results[i].gen_best));
//		fprintf(stderr, "%08x %lf\n", ((su_genome_t*)results[i].gen_best)->luts[0], results[i].fit_best);
	}
//	printf("\n\n");
	fprintf(stderr, "\n%f\n", results[GENS - 1].fit_best);
	ls_dump(stderr, (ls_genome_t*)results[GENS - 1].gen_best, CA_W, CA_H);

	for (int i = 0; i < GENS; i++) {
		free(results[i].gen_best);
	}
	free(results);
}

static void
run_pop(ea_cfg_t *ea_cfg, int pop_size)
{
	ea_result_t *results;
	uni_genome_t *g;

	results = ea_ga(ea_cfg, pop_size, GENS);

	printf("# GA %i\n", pop_size);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf %f %f %f %i\n",
		       i,
		       results[i].fit_best, results[i].fit_avg,
		       results[i].dist_avg, results[i].dist_var,
		       ls_size((ls_genome_t*)results[i].gen_best));
	}
//	printf("\n\n");
	fprintf(stderr, "%f ", results[GENS - 1].fit_best);
	ls_dump(stderr, (ls_genome_t*)results[GENS - 1].gen_best, CA_W, CA_H);

	for (int i = 0; i < GENS; i++) {
		free(results[i].gen_best);
	}
	free(results);
}

static int goal_compact[] = { 0xff, 0xdb, 0xdb, 0xff, 0xa5, 0xbd, 0x81, 0xff };
//static int goal_compact[] = { 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55 };
//static int goal_compact[] = { 0x00, 0x00, 0xfe, 0x7f, 0xe6, 0x67, 0xe6, 0x67, 0xe6, 0x67, 0xe6, 0x67, 0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 0xe6, 0x67, 0xe6, 0x67, 0x06, 0x60, 0x06, 0x60, 0xfe, 0x7f, 0x00, 0x00 };

static int goal[CA_W * CA_H];

static double
ls_fitness(void *g, void *ud)
{
	ls_genome_t *gen = g;
	int array[CA_W * CA_H];
	int correct = 0;

	ls_run(gen, array, CA_W, CA_H);

	for (int x = 0; x < CA_W; x++) {
		for (int y = 0; y < CA_H; y++) {
			int n = ls_xy_to_n(x, y, CA_W, CA_H);

			correct += array[n] == goal[n];
		}
	}

	return correct;
}

int
main(int argc, char *argv[])
{
	ca = ca_new(CA_W, CA_H);
	ls2_cfg.user_data = ca;
	ls2_cfg.calc_fitness = ls_fitness;

	fft_init(WIN_SIZE);

	_srand();

	for (int y = 0; y < CA_H; y++) {
		for (int x = 0; x < CA_W; x++) {
			goal[ls_xy_to_n(x, y, CA_W, CA_H)] = ((goal_compact[y] >> x) & 1);
			fprintf(stderr, "%i %i %i\n", x, y, goal[ls_xy_to_n(x, y, CA_W, CA_H)]);
		}
	}

// 	printf("# Max\n0 %lf\n%i %lf\n\n\n", 384.0 * 1, GENS - 1, 384.0 * 1);

//	fprintf(stderr, "%i %i %i\n", 100, 1, GENS);
	for (int i = CELL_FROM; i <= CELL_TO; i++) {
		ls_types = i;
		printf("# %i\n", i);
		for (int j = 0; j < TIMES; j++) {
//			run_n(&ls_cfg, 4);
			run_pop(&ls2_cfg, POP_SIZE);
			printf("\n");
		}
		printf("\n");
	}

	return 0;
}
