#define _XOPEN_SOURCE
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "ea.h"
#include "ca.h"

#define GENS 10000
#define POP_SIZE 64
#define CA_W 8
#define CA_H 8

#define WIN_SIZE 16


#if 1
//#define FIT_FUNC(o, i) (count_bits((o) ^ (~(i) & 0xff)))
#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))
#define FIT_MAX 8
#else
//#define FIT_FUNC(o, i) ((o) == (~(i) & 0xff))
#define FIT_FUNC(o, i) ((o) == ((i) & 0xff))
#define FIT_MAX 1
#endif

#define GEN
#define GA

#include "util.h"

#define CELL_FROM 2
#define CELL_TO 4
#ifdef GA
#define TIMES 1
#else
#define TIMES 10
#endif

static ca_t *ca;

static void
run_n(ea_cfg_t *ea_cfg, int n)
{
	ea_result_t *results;

	results = ea_one_plus_n(ea_cfg, n, GENS);

	printf("# 1+%i-%i\n", n, cell_count);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf %f %f %f %i\n",
		       i,
		       results[i].fit_best, results[i].fit_avg,
		       results[i].dist_avg, results[i].dist_var,
		       ls_size((ls_genome_t*)results[i].gen_best));
//		fprintf(stderr, "%08x %lf\n", ((su_genome_t*)results[i].gen_best)->luts[0], results[i].fit_best);
	}
//	printf("\n\n");
	fprintf(stderr, "\n%f\n", results[GENS - 1].fit_best);
	ls_dump(stderr, (ls_genome_t*)results[GENS - 1].gen_best, CA_W, CA_H);

	for (int i = 0; i < GENS; i++) {
		free(results[i].gen_best);
	}
	free(results);
}

static void
run_pop(ea_cfg_t *ea_cfg, int pop_size)
{
	static int prefix = 0;
	ea_result_t *results;

	results = ea_ga(ea_cfg, pop_size, GENS);

	printf("# GA %i\n", pop_size);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf %f %f %f %i\n",
		       i,
		       results[i].fit_best, results[i].fit_avg,
		       results[i].dist_avg, results[i].dist_var,
		       ls_size((ls_genome_t*)results[i].gen_best));
	}
//	printf("\n\n");
	fprintf(stderr, "%f ", results[GENS - 1].fit_best);
	ls_dump(stderr, (ls_genome_t*)results[GENS - 1].gen_best, CA_W, CA_H);

	for (int i = 0; i < pop_size; i++) {
		char filename[256];
		FILE *f;

		sprintf(filename, "gens/%i-%02i.gen", prefix, i);

		f = fopen(filename, "w");
		fprintf(f, "%lf\n", results->last_fit[i]);
		ls_dump(f, (ls_genome_t*)results->last_gen[i], CA_W, CA_H);
		fclose(f);

		free(results->last_gen[i]);
	}
	prefix++;

	for (int i = 0; i < GENS; i++) {
		free(results[i].gen_best);
	}
	free(results);
}

static int goal_compact[] = { 0xff, 0xdb, 0xdb, 0xff, 0xa5, 0xbd, 0x81, 0xff };
//static int goal_compact[] = { 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55 };
static int goal[CA_W * CA_H];

#define STEP 11

static int inputs[] =
{
#if CA_H == 7
#define IN_MASK 0x7f
	/*
	0,
	1,8,2,64,16,4,32,
	36,33,20,34,48,3,40,72,
	100,74,41,49,52,42,104,38,
	75,39,51,45,58,54,108,99,
	107,59,122,62,55,118,47,31,
	119,126,63,123,111,95,125,
	127,
	*/
	0,
	2,1,64,8,4,16,32,
	17,40,48,33,20,96,18,72,3,65,12,66,5,6,10,24,80,34,68,9,36,
	26,13,41,76,25,35,52,104,81,67,112,7,56,98,42,37,49,73,19,69,44,22,82,28,70,11,74,84,100,97,21,50,88,14,38,
	113,43,106,120,45,101,78,105,90,57,89,58,85,29,54,60,75,71,92,15,39,27,53,86,108,99,114,102,116,23,77,83,46,51,30,
	121,55,87,61,115,124,79,47,109,31,103,117,62,93,59,91,110,122,107,94,118,
	123,95,111,125,63,126,119,
	127,
#elif CA_H == 8
#define IN_MASK 0xff
	/*
	0,
	16,1,4,128,
	17,132,34,20,
	21,81,22,152,
	62,205,185,218,
	187,126,215,222,
	239,223,254,127,
	255
	*/
	/*
	0,
	16,4,2,8,32,64,128,1,
	10,130,144,18,20,36,6,68,
	84,162,52,134,38,49,67,112,
	171,218,182,234,179,155,59,167,
	63,246,231,243,237,252,207,245,
	247,127,223,191,253,254,239,251,
	255,
	*/
	0,
	16,1,64,2,4,8,32,128,
	96,18,3,24,5,6,72,136,68,144,17,129,192,9,36,80,33,66,48,132,160,34,40,65,12,20,10,130,
	146,112,200,69,14,164,133,152,131,50,162,176,41,38,13,137,148,44,26,88,196,49,67,82,21,56,42,134,11,19,208,37,81,76,74,52,7,22,28,138,98,168,161,104,25,73,194,145,70,193,100,97,224,35,84,140,
	118,236,151,244,227,248,109,218,62,55,233,182,171,211,110,230,31,241,91,214,103,107,47,217,229,234,87,143,124,155,242,188,185,93,213,199,94,181,179,186,158,206,117,174,203,205,167,115,122,173,121,157,79,59,61,220,
	175,187,123,250,222,125,111,219,252,183,159,63,246,238,190,231,249,221,243,237,207,215,126,245,119,189,95,235,
	239,223,191,253,251,127,247,254,
	255,

#elif CA_H == 9
#define IN_MASK 0x1ff
	0,
	4,2,32,16,64,256,128,8,
	384,40,257,12,66,20,96,65,
	42,152,416,261,448,73,100,224,
	294,101,46,170,297,390,198,78,
	121,358,143,122,230,482,339,157,
	437,444,311,350,359,190,411,485,
	508,415,251,445,127,319,381,367,
	479,507,383,509,255,495,503,510,
	511,
#endif
};


static double
ls_fitness(void *g, void *ud)
{
	double ret = 0;
	ca_t *ca = ud;
	ls_genome_t *gen = g;
	uint32_t *out;
	double *avg_specs[2];
	int cluster_size[2];
	int ei = 0;
	int oi = 0;
	int size = sizeof(inputs) / sizeof(int);

	for (int i = 0; i < 2; i++) {
		avg_specs[i] = malloc(sizeof(double) * size / 2);
		cluster_size[i] = size / 2;
	}

	ls_set_luts(ca, gen, CA_W, CA_H);

	for (int i = 0; i < size; i++) {
		int min = count_bits(inputs[i]) < ((CA_H + 1) / 2);

		double fft_in[WIN_SIZE];
		double *spec;

		out = ca_run_initial(ca, inputs[i], WIN_SIZE * 2, NULL, gen->initial);

		for (int i = 0; i < WIN_SIZE; i++) {
			fft_in[i] = count_bits(out[i + WIN_SIZE]);
		}

		spec = fft_run(fft_in);
		avg_specs[min][min ? ei++ : oi++] = spec_avg(spec, WIN_SIZE / 2 + 1);

		free(out);
	}

	ret = dunn_index(avg_specs, 2, cluster_size);
//	ret = berg_index(avg_specs, 2, cluster_size);
//	ret = dunn_index(avg_specs, 2, cluster_size) + berg_index(avg_specs, 2, cluster_size);
	/*
	for (int i = 0; i < size / 2; i++) {
		ret += 8 - avg_specs[0][i];
		ret += avg_specs[1][i] - 1;
	}
	*/

	for (int i = 0; i < 2; i++) {
		free(avg_specs[i]);
	}

	return ret;
}

int
main(int argc, char *argv[])
{
	ca = ca_new(CA_W, CA_H);
	ls_cfg.user_data = ca;
	ls_cfg.calc_fitness = ls_fitness;

	for (int y = 0; y < CA_H; y++) {
		for (int x = 0; x < CA_W; x++) {
			goal[ls_xy_to_n(x, y, CA_W, CA_H)] = ((goal_compact[y] >> x) & 1);
			fprintf(stderr, "%i %i %i\n", x, y, goal[ls_xy_to_n(x, y, CA_W, CA_H)]);
		}
	}

	fft_init(WIN_SIZE);

	_srand();

// 	printf("# Max\n0 %lf\n%i %lf\n\n\n", 384.0 * 1, GENS - 1, 384.0 * 1);

//	fprintf(stderr, "%i %i %i\n", 100, 1, GENS);
	for (int i = CELL_FROM; i <= CELL_TO; i++) {
		ls_types = i;
		printf("# %i\n", i);
		for (int j = 0; j < TIMES; j++) {
#ifdef GA
			run_pop(&ls_cfg, POP_SIZE);
#else
			run_n(&ls_cfg, 4);
#endif
			printf("\n");
		}
		printf("\n");
	}

	return 0;
}
