#define _XOPEN_SOURCE
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "ea.h"
#include "ca.h"

#define GENS 100000
#define CA_W 8
#define CA_H 8

#if 1
//#define FIT_FUNC(o, i) (count_bits((o) ^ (~(i) & 0xff)))
#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))
#define FIT_MAX 8
#else
//#define FIT_FUNC(o, i) ((o) == (~(i) & 0xff))
#define FIT_FUNC(o, i) ((o) == ((i) & 0xff))
#define FIT_MAX 1
#endif

#include "util.h"


static ca_t *ca;

static void
check_ca(nu_genome_t *gen)
{
	ca_set_luts(ca, gen->luts);

	for (uint32_t in = 0; in < 256; in++) {
		uint32_t *outs = ca_run(ca, in, 128, NULL);

		for (int j = 16; j < 128; j++) {
			if (FIT_FUNC(outs[j], in) != FIT_MAX) {
				printf("\t[%i] %08x != %08x\n", j, in, outs[j]);
				break;
			}
		}

		free(outs);
	}
}


static void
run_n(ea_cfg_t *ea_cfg, int n)
{
	ea_result_t *results;
	nu_genome_t *g;

	results = ea_one_plus_n(ea_cfg, n, GENS);

	printf("# 1+%i:\n", n);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
		/*
		g = results[GENS - i].gen_best;
		printf("\t%.0lf\n", results[GENS - i].fit_best);//, g->lut);

		if (results[GENS - i].fit_best >= 1792 * FIT_MAX) {
			check_ca(g);
		}
		*/
	}
	printf("\n\n");

	free(results);
}

static void
run_pop(ea_cfg_t *ea_cfg, int pop_size)
{
	ea_result_t *results;
	nu_genome_t *g;

	results = ea_ga(ea_cfg, pop_size, GENS);

	printf("# GA %i:\n", pop_size);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
		/*
		g = results[GENS - i].gen_best;
		printf("\t%.0lf\n", results[GENS - i].fit_best);//, g->lut);

		if (results[GENS - i].fit_best >= 1792 * FIT_MAX) {
			check_ca(g);
		}
		*/
	}
	printf("\n\n");

	free(results);
}

int
main(int argc, char *argv[])
{
	ca = ca_new(CA_W, CA_H);
	nu_cfg.user_data = ca;

	srand48(time(NULL));

	printf("# Max\n0 %lf\n%i %lf\n\n\n", 1792.0 * FIT_MAX, GENS - 1, 1792.0 * FIT_MAX);

	run_n(&nu_cfg, 1);
	run_n(&nu_cfg, 4);
	run_pop(&nu_cfg, 16);

	return 0;
}
