#define _XOPEN_SOURCE
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "ca.h"
#include "ea.h"

#define GENS 1000
#define CA_W 8
#define CA_H 8

#define WIN_SIZE 16

#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))
#define FIT_MAX 8
#include "util.h"

static ca_t *ca;

static double
fitness(su_genome_t *gen)
{
	lut_t luts[CA_W * CA_H];

	for (int i = 0; i < CA_W * CA_H; i++) {
		luts[i] = gen->luts[i % cell_count];
	}

	ca_set_luts(ca, luts);

	return fft_fitness1(ca, gen->initial);
}

static void
check_roughness()
{
	su_genome_t *gen = su_new();
	double f;
	do {
		free(gen);
		gen = su_new();
		gen->initial = 0;
		f = fitness(gen);
	} while (0); //fabs(f) < 500);


#if 1
	/*
	for (int i = 0; i < 256; i++) {
		gen->initial = i;
		double g = fitness(gen);
		printf("%f\n", (f - g));
	}
	*/

	printf("%f\n", f);
#else
	for (int byte = 0; byte < cell_count; byte++) {
		for (int bit = 0; bit < 32; bit++) {
			su_genome_t *copy = su_copy(gen);
			double g;

			copy->luts[byte] ^= 1 << bit;
			g = fitness(copy);

			printf("%f\n", (f - g));

			free(copy);
		}
	}
	/*
	for (int bit = 0; bit < 8; bit++) {
		su_genome_t *copy = su_copy(gen);
		double g;

		copy->initial ^= 1 << bit;
		g = fitness(copy);

		printf("%f\n", (f - g));

		free(copy);
	}
	*/
#endif
}

int
main(int argc, char *argv[])
{
	ca = ca_new(CA_W, CA_H);
	cell_count = 1;

	fft_init(WIN_SIZE);
	srand48(0);

	for (int i = 0; i < 320000; i++) {
		check_roughness();
	}
	return 0;
}
