#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ca.h"

#include "ea.h"
#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))

#define CA_W 8
#define CA_H 8
#define WIN_SIZE 16

#include "util.h"

#define ITS CA_W
#define STEP 1

static int
xy_to_n(int x, int y, int w, int h)
{
	return x + y * w;
//	return y + x * h;

	/*
	int n = 0;

	x -= (w - 1) / 2;
	y -= h / 2;

	if        ( x > abs(y)) {
		n = 4 * x * x - 3 * x - y;
	} else if (-x > abs(y)) {
		n = 4 * x * x - 1 * x + y;
	} else if ( y >= abs(x)) {
		n = 4 * y * y + 3 * y + x;
	} else if (-y >= abs(x)) {
		n = 4 * y * y + 1 * y - x;
	}

	return n;
	*/
}

int
main(int argc, char *argv[])
{
	su_genome_t *gen;
	int max_input;

	gen = su_read(stdin);
	su_dump(stdout, gen);

	max_input = 1 << CA_H;

	ca_t *ca;
    uint32_t *outs;
    uint8_t *states;
    int mode = 0;
    int w = CA_W;
    int h = CA_H;
	lut_t luts[CA_W * CA_H];

	if (argc > 1) {
		mode = atoi(argv[1]);
	}

	ca_fft_init();
	fft_init(WIN_SIZE);

	FILE *graph = NULL;
	if (argc > 2) {
		graph = fopen(argv[2], "w");
	}

    ca = ca_new(w, h);

	for (int x = 0; x < CA_W; x++) {
		for (int y = 0; y < CA_H; y++) {
			int n = xy_to_n(x, y, CA_W, CA_H);

//			luts[x + y * CA_W] = gen->luts[n % cell_count];
			luts[x + y * CA_W] = gen->luts[x % cell_count];
		}
	}

	ca_set_luts(ca, luts);

	outs = ca_run_initial(ca, 0, WIN_SIZE * 2, &states, gen->initial);
	if (graph != NULL) {
		ca_dump(graph, 0, states, w, h, WIN_SIZE * 2);
	}
	free(outs);
	free(states);

	outs = ca_run_initial(ca, 0xaa, WIN_SIZE * 2, &states, gen->initial);
	if (graph != NULL) {
		ca_dump(graph, 0xaa, states, w, h, WIN_SIZE * 2);
	}
	free(outs);
	free(states);

	outs = ca_run_initial(ca, 0xffffffff, WIN_SIZE * 2, &states, gen->initial);
	if (graph != NULL) {
		ca_dump(graph, 255, states, w, h, WIN_SIZE * 2);
	}
	free(outs);
	free(states);

	fprintf(graph, "\n\n# Data\n");

	for (int i = 0; i < max_input; i ++) {
		double fft_in[WIN_SIZE];
		double *spec;
		int min;

		switch (count_bits(i)) {
		case 0:
		case 1:
		case 2:
		case 3:
			min = 1;
			break;
		case 4:
			min = 2;
			break;
		default:
			min = 0;
			break;
		}

		outs = ca_run_initial(ca, i, WIN_SIZE * 2, NULL, gen->initial);

		if (mode) {
			if (graph != NULL) {
				fprintf(graph, "%i %i %i\n", i, min, outs[WIN_SIZE * 2 - 1]);
			}
		} else {
			for (int i = 0; i < WIN_SIZE; i++) {
				fft_in[i] = count_bits(outs[i + WIN_SIZE]);
			}

			spec = fft_run(fft_in);

			if (graph != NULL) {
				fprintf(graph, "%i %i %f\n", i, min, spec_avg(spec, WIN_SIZE / 2 + 1));
			}
		}

		free(outs);
	}

	if (graph != NULL) {
		fprintf(graph, "\n\n# LS-img\n");
		for (int y = 0; y < h; y++) {
			for (int x = 0; x < w; x++) {
				fprintf(graph, "%i %i %i\n", x, y, xy_to_n(x, y, CA_W, CA_H) % cell_count);
			}
		}

		fclose(graph);
	}

    return 0;
}
