#define _XOPEN_SOURCE
#include <stdlib.h>

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "ea.h"
#include "ca.h"

#define GENS 1000
#define CA_W 8
#define CA_H 8

#if 1
//#define FIT_FUNC(o, i) (count_bits((o) ^ (~(i) & 0xff)))
#define FIT_FUNC(o, i) (count_bits((o) ^ ((i) & 0xff)))
#define FIT_MAX 8
#else
//#define FIT_FUNC(o, i) ((o) == (~(i) & 0xff))
#define FIT_FUNC(o, i) ((o) == ((i) & 0xff))
#define FIT_MAX 1
#endif

#include "util.h"


static ca_t *ca;

static void
check_ca(nu_genome_t *gen)
{
	ca_set_luts(ca, gen->luts);

	for (uint32_t in = 0; in < 256; in++) {
		uint32_t *outs = ca_run(ca, in, 128, NULL);

		for (int j = 16; j < 128; j++) {
			if (FIT_FUNC(outs[j], in) != FIT_MAX) {
				printf("\t[%i] %08x != %08x\n", j, in, outs[j]);
				break;
			}
		}

		free(outs);
	}
}

static nu_genome_t *wide_best = NULL;

static void
run_n(ea_cfg_t *ea_cfg, int w, int n)
{
	ea_result_t *results;
	nu_genome_t *g;

	results = ea_one_plus_n(ea_cfg, n, GENS);

	printf("# %i:\n", w);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
		/*
		g = results[GENS - i].gen_best;
		printf("\t%.0lf\n", results[GENS - i].fit_best);//, g->lut);

		if (results[GENS - i].fit_best >= 1792 * FIT_MAX) {
			check_ca(g);
		}
		*/
	}
	printf("\n\n");

	wide_best = nu_copy(results[GENS - 1].gen_best);
	free(results);
}

static void
run_pop(ea_cfg_t *ea_cfg, int pop_size)
{
	ea_result_t *results;
	nu_genome_t *g;

	results = ea_ga(ea_cfg, pop_size, GENS);

	printf("# GA %i:\n", pop_size);
	for (int i = 0; i < GENS; i++) {
		printf("%i %lf\n", i, results[i].fit_best);
		/*
		g = results[GENS - i].gen_best;
		printf("\t%.0lf\n", results[GENS - i].fit_best);//, g->lut);

		if (results[GENS - i].fit_best >= 1792 * FIT_MAX) {
			check_ca(g);
		}
		*/
	}
	printf("\n\n");

	free(results);
}

static int width = 0;

static double
wide_fitness(void *g, void *ud)
{
	ca_t *ca = ud;
	nu_genome_t *gen = g;
	double tot = 0;
	uint32_t luts[CA_H * width];

	for (int x = 0; x < width; x++) {
		for (int y = 0; y < CA_H; y++) {
			luts[x + y * width] = gen->luts[x + y * CA_W];
		}
	}

	ca_set_luts(ca, luts);

	for (int i = 0; i < 16; i++) {
		uint32_t in = i * 15;// + _rand() % 16;
		//uint32_t in = _rand() & 0xff;
		uint32_t *outs = ca_run(ca, in, 128, NULL);

		for (int j = 16; j < 128; j++) {
			tot += FIT_FUNC(outs[j], in);
			/*
			if (FIT_FUNC(outs[j], in)) {
				tot++;
			}
			*/
		}

		free(outs);
	}

	return tot;
}

static void *
wide_new(void)
{
	return wide_best;
}

int
main(int argc, char *argv[])
{
	nu_cfg.calc_fitness = wide_fitness;
	nu_cfg.new = wide_new;

	wide_best = nu_new();

	srand48(time(NULL));

	printf("# Max\n0 %lf\n%i %lf\n\n\n", 1792.0 * FIT_MAX, GENS - 1, 1792.0 * FIT_MAX);

	for (int w = 1; w <= 8; w++) {
		width = w;
		ca = ca_new(w, CA_H);
		nu_cfg.user_data = ca;
		run_n(&nu_cfg, w, 4);
	}

	return 0;
}
